/*      
** avl_tree.h	AVL Balanced Tree Library Header
**
** Copyright (c) 1996  Hughes Technologies Pty Ltd
**
*************************************************************************
**
*/

#ifndef __AVL_TREE_H
#define __AVL_TREE_H


/* 
** AVL Tree Dump Options 
*/
#define AVL_QUIET	0
#define	AVL_VERBOSE	1
#define AVL_PATH_LEN    256


/*
** Support for Version 2 format trees with reverse lookup deletion support
*/
#define	AVL_V1		1
#define AVL_V2		2



typedef struct {
	int	magic;		/* Magic number for tree (for testing) */
	off_t	fileLen;	/* Has another proc extended the file?  */
} ptr_sbk;


typedef struct {
	int	magic,		/* Magic number for tree (for testing) */
		keyLen,		/* Length of key values */
		nodeLen,	/* On disk node length */
		keyType,	/* Data type  of key */
		flags;		/* Flags for unique etc. */
	int	root,		/* Node number of tree root */
		freeList;	/* Node number of start of free list */
	off_t	fileLen;	/* Has another proc extended the file?  */
	u_int	numEntries,	/* Total number of entries in tree */
		numKeys;	/* Number of distinct key values */
} avl_sbk;


typedef struct {
	char	*treeMap,	/* Pointer to mapping of the file */
		*ptrMap,	/* Pointer to mapping of the ptr file */
		*memTree,	/* Pointer to malloc space for in-core tree*/
		path[AVL_PATH_LEN + 1];
	off_t	treeMapLen,	/* Length of file mapping */
		ptrMapLen,	/* Length of the ptr file mapping */
		memLen;		/* Size of malloced region of in-core tree */
	avl_sbk	*sblk;		/* Pointer to file superblock */
	ptr_sbk *psblk;		/* Pointer to ptr file superblock */
	int	treeFd,		/* Open file descriptor of tree file */
		ptrFd,		/* Open file descriptor for ptr file */
		remapped,	/* Flag to indicate the tree was remapped */
		curNode,	/* Flags last node of last tree search */
		version;	/* AVL format version, calculated from magic */
} avltree;



typedef struct {
	int	curNode,	/* Current node for get/next */
		curDup,		/* Current dup for get/next */
		nextDup;	/* So we can track if the dup chain changed */
} avl_cur;


typedef struct node_s {
	int	nodeNum,	/* Node number for this node */
		left,		/* Node number of left node */
		right,		/* Node number of right node */
		parent,		/* Node number of parent node */
		dup;		/* Node number of head of dup list */
	short	height;		/* Node height indicator */
	char	*key;		/* Pointer to key value */
	off_t	data;		/* Offset value for main data table */
} avl_nod;



/* Prototypes */

int 	avlCreate();
int 	avlInsert();
int 	avlDelete();
int 	avlTestTree();
void 	avlClose();
void    avlSync();
void	avlDumpTree();
void	avlSetCursor();
void	avlPrintElement();
avltree	*avlOpen();
avltree	*avlMemCreate();
/*avl_nod *avlGetNode();*/
avl_nod	*avlFindNode();
avl_nod	*avlGetFirst();
avl_nod	*avlGetLast();
avl_nod	*avlGetNext();
avl_nod	*avlGetPrev();
avl_nod	*avlLookup();


#define avlKey(n) ((char*)n+sizeof(avl_nod))


#endif
